// David Denny, 1001915603

// Modular Exponent Subtractor -- Five-Bit Variation
module ExponentSubtractorFiveBit (
	input		[4:0] A, 	// Inputs
	input		[4:0] B,
	output	[4:0] R,		// Outputs
	output	Cout);
	
	// Internal wires
	logic		[4:0] SubtractResult;
	logic		[4:0] TwosCompResult;
	
	// Ripple Carry Subtractor Instantiation
	RippleCarrySubtractor #(5) subtractor
	(
		.A(A),
		.B(B),
		.R(SubtractResult),
		.Cout(Cout)
	);
	
	// Two's Complement Converter Instantiation
	twosCompConverter #(5) twosComp
	(
		.in(SubtractResult),
		.out(TwosCompResult)
	);
	
	// Modular Multiplexer Instantiation
	modularMux #(5) muxResult
	(
		.R(SubtractResult),
		.B(TwosCompResult),
		.Cout(Cout),
		.out(R)
	);

endmodule

// Ripple Carry Subtractor
module RippleCarrySubtractor #(parameter N = 5)(
	input [N-1:0] A, B, 	// Declare input ports
	output [N-1:0] R, 	// Declare output ports for Sum
	output Cout); 			// Declare Carry-Out port
	
	wire [N:0] C; // Declare internal nets
	assign C[0] = 1'b1; // Assign 1 to least significant carry-in for subtraction
	wire [N-1:0] Bin;
		
	// Instantiate the full adder module for each stage of the ripple carry subtractor
	genvar i;
		generate
			for (i = 0; i < N; i = i + 1) begin : fullSubtractors
				xor (Bin[i], 1'b1, B[i]); // One's complement of input B
				FAbehave full_adder (
					.ai(A[i]),
					.bi(Bin[i]),
					.cini(C[i]),
					.si(R[i]),
					.couti(C[i+1])
				);
			end
	endgenerate
		
	assign Cout = C[N]; // Carry-Out of the last stage, to be used in multiplexer
		
endmodule

// Two's Complement Converter
module twosCompConverter #(parameter N = 5)(
    input  logic [N-1:0] in,   	// 5-bit input
    output logic [N-1:0] out);	// 5-bit output

	// All needed wires for converter
	logic [N-1:0] invertedIn;   // Inverted input
	logic [N-1:0] sum;          // Sum of each half adder
	logic [N-1:0] carry;        // Carry of each half adder
	
	// XOR input with 1 to get one's complement
	genvar j;
		generate
			for (j = 0; j < N; j = j + 1) begin : fullSubtractors
				xor (invertedIn[j], 1'b1, in[j]); // One's complement on input B
			end
	endgenerate

	// Adding 1 to one's complement using Half Adders
	// Sum register will then be the two's complement of the input
	halfAdder HalfAdderBit0 (
		.A(invertedIn[0]), 
		.B(1'b1), // Adding 1 to LSB with first half adder
		.Sum(sum[0]), 
		.Cout(carry[0])
	);

	// Generate the other half adders for the rest of the bits
	genvar i;
	generate
		for (i = 1; i < N; i = i + 1) begin : halfAdders
			halfAdder HalfAdderOtherBits (
				.A(invertedIn[i]), 
				.B(carry[i-1]),
				.Sum(sum[i]), 
				.Cout(carry[i])
			);
		end
	endgenerate

	// Output the two's complement of the input
	assign out = sum;

endmodule

// Half Adder Behavior
module halfAdder (
	input A, B,
	output reg Sum, Cout);
		always @ (A,B)
			case ({A,B})
				2'b00: begin Cout = 1'b0; Sum = 1'b0; end
				2'b01: begin Cout = 1'b0; Sum = 1'b1; end
				2'b10: begin Cout = 1'b0; Sum = 1'b1; end
				2'b11: begin Cout = 1'b1; Sum = 1'b0; end
			endcase
endmodule

// Full Adder Behavior
module FAbehave (
	input ai, bi, cini,
	output reg si, couti);
		always @ (ai,bi,cini)
			case ({ai,bi,cini})
				3'b000: begin couti = 1'b0; si = 1'b0; end
				3'b001: begin couti = 1'b0; si = 1'b1; end
				3'b010: begin couti = 1'b0; si = 1'b1; end
				3'b011: begin couti = 1'b1; si = 1'b0; end
				3'b100: begin couti = 1'b0; si = 1'b1; end
				3'b101: begin couti = 1'b1; si = 1'b0; end
				3'b110: begin couti = 1'b1; si = 1'b0; end
				3'b111: begin couti = 1'b1; si = 1'b1; end
			endcase
endmodule
