// David Denny, 1001915603

// 16-bit Floating-Point Adder
module LCD_Board_Test_LCD
	 (
		input                clk          , // 50MHz Clock
		input                lcd_reset    , // LCD Reset
		inout    [7:0]       lcd_data     , // LCD Data Bus
		output               lcd_rs       , // LCD Register Select
		output               lcd_rw       , // LCD Read Write Select
		output               lcd_e        , // LCD Execute

		// Keypad I/O (clk and lcd_reset is shared between LCD and Keypad)
		input [3:0] row, // input row/col
		input Ain, Bin, Rin,
		output [3:0] col,
		output [7:0] out,
		output [7:0] out2,
		output [7:0] out3,
		output [7:0] out4,
		output Aled, Bled, Rled,
		
		// Condition Code Indicators
		output Overflow, Underflow
	);   
	
	// ==================== All Internal Wires ====================
	  
	// Internal Wire used for Keypad
	wire [15:0] keypadWire;
	wire [15:0] Aout;
	wire [15:0] Bout;
	wire [15:0] Rout;
	assign Aled = Ain; // Used to output A to LCD and registers	- Just an indicator light
	assign Bled = Bin; // Used to output B to LCD and registers	- Just an indicator light
	assign Rled = Rin; // Used to output R to LCD and registers - Just an indicator light
	
	// More Internal Wires for Addition Arithmetic
	wire signA, signB, expSubCout;
	wire [4:0] exponentA, exponentB, expSubResult, exponentIncMuxOut, exponentIncrementOut;
	wire [9:0] mantissaA, mantissaB, mantissaShifterMuxOut, mantissaAdderMuxOut, finalMantissa;
	wire [10:0] shiftedMantissa, normalizerMantissa;
	wire [11:0] CLAout;
	
	// Condition Code Wires
	wire Exception, Zero;
	
	// ===================== Input Registers =====================
	
	// Register A
	NBitRegister #(16) regA // 16-bit register
	(
		.D(keypadWire), // Register A input
		.CLK(Ain), // Register A load
		.CLR(lcd_reset), // Register A clear
		.Q(Aout) // Register A output
	);	
	
	assign signA = Aout[15];
	assign exponentA = Aout[14:10];
	assign mantissaA = Aout[9:0];

	// Register B
	NBitRegister #(16) regB // 16-bit register
	(
		.D(keypadWire), // Register B input
		.CLK(Bin), // Register B load
		.CLR(lcd_reset), // Register B clear
		.Q(Bout) // Register B output
	);
	
	assign signB = Bout[15];
	assign exponentB = Bout[14:10];
	assign mantissaB = Bout[9:0];
	
	// ========== All Adder Component Module Instantiations ==========
	
	// Exponent Subtractor Instantiation
	ExponentSubtractorFiveBit expSub // 5-bit Exponent Subtractor
	(
		.A(exponentA), // Exponent A Input
		.B(exponentB), // Exponent B Input
		.R(expSubResult), // Result of Exponent Subtractor
		.Cout(expSubCout) // Register B output
	);
	
	// Modular Multiplexer Instantiation
	modularMux #(10) MantissaShifterMuxInst
	(
		.R(mantissaA),
		.B(mantissaB),
		.Cout(~expSubCout),
		.out(mantissaShifterMuxOut)
	);
	
	// Modular Multiplexer Instantiation
	modularMux #(10) MantissaAdderMuxInst
	(
		.R(mantissaA),
		.B(mantissaB),
		.Cout(expSubCout),
		.out(mantissaAdderMuxOut)
	);
	
	// Modular Multiplexer Instantiation
	modularMux #(5) ExponentIncMuxInst
	(
		.R(exponentA),
		.B(exponentB),
		.Cout(expSubCout),
		.out(exponentIncMuxOut)
	);
	
	// Right Shifter Instantiation
	ElevenBitRightShifter shifter_inst
	(
		.in({1'b1, mantissaShifterMuxOut}),
		.ctrl(expSubResult[3:0]),
		.out(shiftedMantissa) 
	);
	
	// Carry Look-Ahead Adder Instantiation
	TwelveBitGCLAadder CLAadder_inst
	(
		.Aout({1'b0, shiftedMantissa}),
		.Bout({2'b01, mantissaAdderMuxOut}),
		.R(CLAout) 
	);

	// Final Right Shifter Instantiation
	ElevenBitRightShifter final_shifter_inst
	(
		.in(CLAout[10:0]),
		.ctrl({3'b000, CLAout[11]}),
		.out(normalizerMantissa) 
	);
	
	assign finalMantissa = normalizerMantissa[9:0];
	
	// Controlled Exponent Incrementer Instantiation
	FiveBitIncrementer expIncFinal_inst
	(
		.A(exponentIncMuxOut),
		.select(CLAout[11]),
		.S(exponentIncrementOut) 
	);
	
	// Register R
	NBitRegister #(16) regR // 16-bit register
	(
		.D({1'b0, exponentIncrementOut, finalMantissa}), // Register R input
		.CLK(Rin), // Register R load
		.CLR(lcd_reset), // Register R clear
		.Q(Rout) // Floating Point Addition Result
	);
	
	// =================== Condition Code Logic ===================
	
	// This logic is simply taken from Lecture 9.1
	assign Exception = (&exponentA) | (&exponentB);
	assign Zero = Exception ? 1'b0 : (finalMantissa == 10'd0) ? 1'b1 : 1'b0;
	assign Overflow = ((exponentIncrementOut[4] & !exponentIncrementOut[3]) & !Zero);
	assign Underflow = ((exponentIncrementOut[4] & exponentIncrementOut[3]) & !Zero) ? 1'b1 : 1'b0;
	
	// =========== LCD and Keypad Module Instantiations ===========
	
	// LCD instantiation -- LAB 8
	LCD #(
		 .WIDTH(64),
		 .DIGITS(16),
		 .FLOAT(0),
		 .MODE(1),
		 .LINES(4),
		 .CHARS(20),
		 .LINE_STARTS({7'h00, 7'h40, 7'h14, 7'h54})
		 )(
		 .clk(clk),
		 .lcd_data(lcd_data),
		 .lcd_rs(lcd_rs),
		 .lcd_rw(lcd_rw),
		 .lcd_e(lcd_e),
		 .lcd_reset(!lcd_reset),
		 .A(Aout),
		 .B(Bout),
		 .C(Rout),
		 .Operation(2'b01) // PLUS
	);
	
	// Keypad Instantiation -- LAB 2
	InputUnit InputUnit_inst
	(
		.row(row),
		.clk(clk),
		.clear(lcd_reset),
		.col(col),
		.out(out),
		.out2(out2),
		.out3(out3),
		.out4(out4),
		.keypadWire(keypadWire)
	);

endmodule
