// David Denny, 1001915603

// Keypad Module Instantiations
module clock_div
#(
parameter WIDTH = 32,
parameter DIV = 50 // Defaults to 1MHz
)
(
		input clk, reset,
		output clk_out
);
reg [WIDTH-1:0] r_reg;
wire [WIDTH-1:0] r_nxt;
reg clk_track;
	always @(posedge clk or posedge reset)
		begin
			if (reset)
			begin
				r_reg <= 0;
				clk_track <= 1'b0;
			end
			else if (r_nxt == DIV)
			begin
				r_reg <= 0;
				clk_track <= ~clk_track;
			end
			else
					r_reg <= r_nxt;
			end
			assign r_nxt = r_reg+1;
			assign clk_out = clk_track;
endmodule

module keypad_base( // Depends: clock_div, keypad_fsm, keypad_decoder
		input clk,
		input [3:0] row,
		output [3:0] col,
		output [3:0] value,
		output valid,
		// Debug
		output slow_clock,
		output sense,
		output valid_digit,
		output [3:0] inv_row
);
		assign inv_row = ~row;
		
		clock_div #(.DIV(100000)) keypad_clock_divider(
				.clk(clk),
				.clk_out(slow_clock)
		);
		keypad_fsm keypad_fsm(
				.clk(slow_clock),
				.row(inv_row),
				.col(col),
				.sense(sense)
		);
		keypad_decoder #(.BASE(16)) keypad_key_decoder(
				.row(inv_row),
				.col(col),
				.value(value),
				.valid(valid_digit)
		);
		assign valid = valid_digit && sense;
endmodule

module keypad_decoder
#(
	parameter BASE = 10 // Default Base 10 (Options: 10,16)
)
(
	input [3:0] row,
	input [3:0] col,
	output reg [3:0] value,
	output reg valid // Optional
);
always @ (row, col)
begin
	if (BASE == 10)
		begin
			// 1 2 3 A
			// 4 5 6 B
			// 7 8 9 C
			// E 0 F D
			case ({row, col})
				8'b0001_0001: begin value = 1; valid = 1; end // 1
				8'b0001_0010: begin value = 2; valid = 1; end // 2
				8'b0001_0100: begin value = 3; valid = 1; end // 3
				8'b0001_1000: begin value = 10; valid = 1; end // A
				8'b0010_0001: begin value = 4; valid = 1; end // 4
				8'b0010_0010: begin value = 5; valid = 1; end // 5
				8'b0010_0100: begin value = 6; valid = 1; end // 6
				8'b0010_1000: begin value = 11; valid = 1; end // B
				8'b0100_0001: begin value = 7; valid = 1; end // 7
				8'b0100_0010: begin value = 8; valid = 1; end // 8
				8'b0100_0100: begin value = 9; valid = 1; end // 9
				8'b0100_1000: begin value = 12; valid = 1; end // C
				8'b1000_0001: begin value = 14; valid = 1; end // E
				8'b1000_0010: begin value = 0; valid = 1; end // 0
				8'b1000_0100: begin value = 15; valid = 1; end // F
				8'b1000_1000: begin value = 13; valid = 1; end // D
				default: begin value = 0; valid = 0; end // Misc
			endcase
		end
	else if (BASE == 16)
		begin
			// 0 1 2 3
			// 4 5 6 7
			// 8 9 A B
			// C D E F
			case ({row, col})
				8'b0001_0001: begin value = 0; valid = 1; end // 0
				8'b0001_0010: begin value = 1; valid = 1; end // 1
				8'b0001_0100: begin value = 2; valid = 1; end // 2
				8'b0001_1000: begin value = 3; valid = 1; end // 3
				8'b0010_0001: begin value = 4; valid = 1; end // 4
				8'b0010_0010: begin value = 5; valid = 1; end // 5
				8'b0010_0100: begin value = 6; valid = 1; end // 6
				8'b0010_1000: begin value = 7; valid = 1; end // 7
				8'b0100_0001: begin value = 8; valid = 1; end // 8
				8'b0100_0010: begin value = 9; valid = 1; end // 9
				8'b0100_0100: begin value = 10; valid = 1; end // A
				8'b0100_1000: begin value = 11; valid = 1; end // B
				8'b1000_0001: begin value = 12; valid = 1; end // C
				8'b1000_0010: begin value = 13; valid = 1; end // D
				8'b1000_0100: begin value = 14; valid = 1; end // E
				8'b1000_1000: begin value = 15; valid = 1; end // F
				default: begin value = 0; valid = 0; end // Misc
			endcase
		end
	else
		begin
			value = 0; valid = 0; // Invalid Base
		end
	end
endmodule

module keypad_fsm(
	input clk,
	input [3:0] row,
	output reg [3:0] col,
	output sense,
// Debug
	output reg [3:0] state,
	output trig
);
	assign trig = row[0] || row[1] || row[2] || row[3];
	assign sense = state == 10;
	always @ (posedge clk)
		begin
			case (state)
				0: begin col = 4'b1111; state = 1; end
				1: if (trig) begin state = 2; end
				2: begin col = 4'b0001; state = 3; end
				3: if (trig) begin state = 10; end else begin state = 4; end
				4: begin col = 4'b0010; state = 5; end
				5: if (trig) begin state = 10; end else begin state = 6; end
				6: begin col = 4'b0100; state = 7; end
				7: if (trig) begin state = 10; end else begin state = 8; end
				8: begin col = 4'b1000; state = 9; end
				9: if (trig) begin state = 10; end else begin state = 0; end
				10: begin state = 11; end
				11: if (~trig) begin state = 0; end
			endcase
		end
endmodule

module keypad_input // Depends: keypad_base(clock_div, keypad_fsm, keypad_decoder), shift_reg
#( parameter DIGITS = 4
)
(		input clk,
		input reset,
		input [3:0] row,
		output [3:0] col,
		output [(DIGITS*4)-1:0] out,
		// Debug
		output [3:0] value,
		output trig
);
		keypad_base keypad_base(
				.clk(clk),
				.row(row),
				.col(col),
				.value(value),
				.valid(trig)
		);
		shift_reg #(.COUNT(DIGITS)) shift_reg(
				.trig(trig),
				.in(value),
				.out(out),
				.reset(reset)
		);
endmodule

module shift_reg
#(
parameter COUNT = 4,
parameter WIDTH = 4
)
(
	input trig, reset, dir, // Left = 0, Right = 1
	input [WIDTH-1:0] in,
	output reg [(COUNT*WIDTH)-1:0] out
);
	always @ (posedge trig, negedge reset)
	begin
		if (~reset)
			out <= 0;
		else
		begin
			if (dir)
			begin // 1 = Right
				out <= out >> WIDTH;
				out[(COUNT*WIDTH)-1:((COUNT*WIDTH)-1)-WIDTH] <= in;
			end
			else
			begin // 0 = Left
				out <= out << WIDTH;
				out[WIDTH-1:0] <= in;
			end
		end
	end
endmodule